#!/usr/bin/env python3

# Automated Simulation of Metal FFF 3D Printing Debinding Process.
# Copyright (C) 2021  Lehrstuhl für Konstruktionslehre und CAD, Universität Bayreuth
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import argparse
import copy
import subprocess
import textwrap
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Tuple

import gmsh
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyvista as pv
import vtk
from scipy.spatial.transform import Rotation


@dataclass
class Z88IntOrderSet:
    elements: Tuple[int, int]
    int_order_displacement: int
    int_order_stress: int


@dataclass
class Z88MaterialSet:
    elements: Tuple[int, int]
    youngs_modulus: float
    poissons_ratio: float


@dataclass
class Z88SPC:
    node: int
    dof: int
    value: float


@dataclass
class Z88Load:
    node: int
    dof: int
    value: float


class Z88ElementType(IntEnum):
    Tet4 = 17
    Tet10 = 16


class Z88Project:
    dimension: int
    nodes: pd.DataFrame
    elements: pd.DataFrame
    spcs: List[Z88SPC]
    loads: List[Z88Load]
    material_sets: List[Z88MaterialSet]
    int_order_sets: List[Z88IntOrderSet]

    def __init__(self) -> None:
        self.dimension = 3
        self.nodes = pd.DataFrame(columns=["dof", "x", "y", "z"])
        self.elements = pd.DataFrame(columns=["type", "nodes"])
        self.spcs = []
        self.loads = []
        self.material_sets = []
        self.int_order_sets = []

    @classmethod
    def from_gmsh(
        cls,
        input_file: Path,
        element_size: float,
        element_order: int = 1,
        gmsh_algorithm: int = 6,
        gmsh_verbose: bool = False,
    ) -> "Z88Project":
        if not input_file.is_file():
            raise ValueError("Input file is no valid file!")

        self = cls()

        gmsh.initialize()
        gmsh.option.setNumber("General.Terminal", int(gmsh_verbose))
        gmsh.option.setNumber("Mesh.Algorithm", gmsh_algorithm)
        gmsh.option.setNumber("Mesh.ElementOrder", element_order)
        gmsh.option.setNumber("Mesh.CharacteristicLengthMin", element_size)
        gmsh.option.setNumber("Mesh.CharacteristicLengthMax", element_size)
        gmsh.option.setNumber("Mesh.HighOrderOptimize", 1)

        file_format = input_file.suffix.lower()
        if file_format in (".stl",):
            # https://gitlab.onelab.info/gmsh/gmsh/-/blob/gmsh_4_7_1/demos/api/remesh_stl.py
            gmsh.merge(str(input_file))
            gmsh.model.mesh.classifySurfaces(40.0 * np.pi / 180.0, True, True)
            gmsh.model.mesh.createGeometry()

            sl = gmsh.model.geo.addSurfaceLoop(
                [e[1] for e in gmsh.model.getEntities(2)]
            )
            v = [gmsh.model.geo.addVolume([sl])]
        elif file_format in (".step", ".stp"):
            gmsh.merge(str(input_file))
            v = [e[1] for e in gmsh.model.getEntities(3)]
        else:
            raise ValueError(f"Unsupported file extension '{file_format}'!")

        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()

        pg = gmsh.model.addPhysicalGroup(3, v)
        gmsh.model.mesh.generate(3)

        node_tags, coordinates = gmsh.model.mesh.getNodesForPhysicalGroup(3, pg)
        coordinates = coordinates.reshape(coordinates.shape[0] // 3, 3)
        self.nodes = pd.DataFrame(
            {
                "dof": np.full(coordinates.shape[0], 3, dtype=np.int64),
                "x": coordinates[:, 0],
                "y": coordinates[:, 1],
                "z": coordinates[:, 2],
            }
        )
        node_mapping = pd.Series(self.nodes.index, index=node_tags)

        tags = gmsh.model.getEntitiesForPhysicalGroup(3, pg)
        for tag in tags:
            element_types, element_tags, node_tags = gmsh.model.mesh.getElements(3, tag)
            for element_type, element_tag, node_tag in zip(
                element_types, element_tags, node_tags
            ):
                node_tag = node_mapping[node_tag].values

                if element_type == 4:
                    element_type = Z88ElementType.Tet4
                    element_nodes = node_tag.reshape(node_tag.shape[0] // 4, 4)[
                        :, [0, 1, 2, 3]
                    ]
                elif element_type == 11:
                    element_type = Z88ElementType.Tet10
                    element_nodes = node_tag.reshape(node_tag.shape[0] // 10, 10)[
                        :, [0, 1, 2, 3, 4, 5, 6, 9, 8, 7]
                    ]
                else:
                    continue

                self.elements = self.elements.append(
                    pd.DataFrame(
                        {
                            "type": np.full(element_tag.shape, element_type),
                            "nodes": (nodes for nodes in element_nodes),
                        }
                    )
                )

        gmsh.finalize()

        return self

    @property
    def n_nodes(self) -> int:
        return len(self.nodes)

    @property
    def n_elements(self) -> int:
        return len(self.elements)

    @property
    def n_dof(self) -> int:
        return self.nodes.dof.sum()

    def transform_mesh(self, matrix: np.array) -> None:
        self.nodes[["x", "y", "z"]] = self.nodes[["x", "y", "z"]].to_numpy() @ matrix

    def assign_material(self, youngs_modulus: float, poissons_ratio: float) -> None:
        self.material_sets = [
            Z88MaterialSet(
                elements=(1, self.n_elements),
                youngs_modulus=youngs_modulus,
                poissons_ratio=poissons_ratio,
            )
        ]

    def assign_integration_order(
        self, int_order_displacement: int, int_order_stress: int
    ) -> None:
        self.int_order_sets = [
            Z88IntOrderSet(
                elements=(1, self.n_elements),
                int_order_displacement=int_order_displacement,
                int_order_stress=int_order_stress,
            )
        ]

    def assign_volume_force(self, density: float, acceleration: float) -> None:
        nodes = self.nodes[["x", "y", "z"]].to_numpy()
        volume_per_node = np.zeros(self.n_nodes)

        def volume_tetrahedron(tet_nodes: np.array) -> float:
            directions = tet_nodes - tet_nodes[0]
            return np.abs(
                (1.0 / 6.0)
                * np.dot(np.cross(directions[1, :], directions[2, :]), directions[3, :])
            )

        def volume_decomposition(tet_nodes: np.array, volume: float) -> np.array:
            distance_center = np.linalg.norm(tet_nodes - tet_nodes.mean(axis=0), axis=1)
            return volume * ((1.0 / distance_center) / np.sum(1.0 / distance_center))

        for element in self.elements[
            self.elements.type == Z88ElementType.Tet4.value
        ].itertuples():
            element_nodes = nodes[element.nodes]
            element_volume = volume_tetrahedron(element_nodes)
            volume_per_node[element.nodes] += volume_decomposition(
                element_nodes, element_volume
            )

        for element in self.elements[
            self.elements.type == Z88ElementType.Tet10.value
        ].itertuples():
            element_nodes = nodes[element.nodes]
            element_volume = (
                volume_tetrahedron(element_nodes[[0, 4, 6, 9]])
                + volume_tetrahedron(element_nodes[[1, 4, 7, 5]])
                + volume_tetrahedron(element_nodes[[2, 8, 6, 5]])
                + volume_tetrahedron(element_nodes[[3, 7, 8, 9]])
                + volume_tetrahedron(element_nodes[[7, 4, 8, 5]])
                + volume_tetrahedron(element_nodes[[4, 8, 6, 5]])
                + volume_tetrahedron(element_nodes[[4, 7, 9, 8]])
                + volume_tetrahedron(element_nodes[[4, 8, 6, 9]])
            )
            volume_per_node[element.nodes] += volume_decomposition(
                element_nodes, element_volume
            )

        force_per_node = volume_per_node * density * acceleration
        for node, load in enumerate(force_per_node):
            self.loads.append(Z88Load(node=node, dof=3, value=-load))

    def write_z88i1(self, path: Path) -> None:
        with path.open("w") as f:
            f.write(
                f"{self.dimension} {self.n_nodes} {self.n_elements} {self.n_dof} 0\n"
            )

            for node in self.nodes.itertuples():
                f.write(
                    f"{node.Index + 1} {node.dof} {node.x:+.15e} {node.y:+.15e} {node.z:+.15e}\n"
                )

            for element in self.elements.itertuples():
                f.write(f"{element.Index + 1} {element.type}\n")
                f.write(f"{' '.join(str(node) for node in element.nodes + 1)}\n")

    def write_z88i2(self, path: Path) -> None:
        with path.open("w") as f:
            f.write(f"{len(self.spcs) + len(self.loads)}\n")

            for spc in self.spcs:
                f.write(f"{spc.node + 1} {spc.dof} 2 {spc.value:+.15e}\n")

            for load in self.loads:
                f.write(f"{load.node + 1} {load.dof} 1 {load.value:+.15e}\n")

    @staticmethod
    def write_z88i5(path: Path) -> None:
        with path.open("w") as f:
            f.write("0\n")

    @staticmethod
    def write_z88elp(path: Path) -> None:
        with path.open("w") as f:
            f.write("0\n")

    def write_z88int(self, path: Path) -> None:
        with path.open("w") as f:
            f.write(f"{len(self.int_order_sets)}\n")

            for int_order_set in self.int_order_sets:
                f.write(
                    f"{int_order_set.elements[0]} {int_order_set.elements[1]} {int_order_set.int_order_displacement} {int_order_set.int_order_stress}\n"
                )

    def write_z88mat(self, path: Path) -> None:
        with path.open("w") as f:
            f.write(f"{len(self.material_sets)}\n")

            for i, material_set in enumerate(self.material_sets):
                material_file = path.parents[0] / f"z88mat_{i}.txt"
                with material_file.open("w") as mf:
                    mf.write(
                        f"{material_set.youngs_modulus} {material_set.poissons_ratio}\n"
                    )

                f.write(
                    f"{material_set.elements[0]} {material_set.elements[1]} {material_file.name}\n"
                )

    @staticmethod
    def write_z88man(path: Path) -> None:
        with path.open("w") as f:
            f.write(
                textwrap.dedent(
                    """\
                DYNAMIC START

                GLOBAL START
                    IBFLAG      0
                    IPFLAG      0
                    IHFLAG      0
                GLOBAL END

                SOLVER START
                    MAXIT       10000
                    EPS         1e-007
                    RALPHA      0.0001
                    ROMEGA      1.1
                SOLVER END

                STRESS START
                    KDFLAG      0
                    ISFLAG      0
                STRESS END

                DYNAMIC END
            """
                )
            )

    def write_z88dyn(self, path: Path) -> None:
        with path.open("w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                DYNAMIC START

                ENGLISH

                COMMON START
                    MAXGS       {(self.n_dof * self.n_dof) // 100}
                    MAXKOI      {self.n_elements * 20}
                    MAXK        {self.n_nodes}
                    MAXE        {self.n_elements}
                    MAXNFG      {self.n_dof}
                    MAXMAT      {len(self.material_sets)}
                    MAXJNT      {len(self.int_order_sets)}
                    MAXPEL      {1}
                    MAXPR       {1}
                    MAXRBD      {len(self.loads) + len(self.spcs)}
                    MAXIEZ      {(self.n_dof * self.n_dof) // 100}
                    MAXGP       {self.n_elements * 5}
                COMMON END

                DYNAMIC END
            """
                )
            )

    def write(self, path: Path) -> None:
        self.write_z88i1(path / "z88i1.txt")
        self.write_z88i2(path / "z88i2.txt")
        self.write_z88i5(path / "z88i5.txt")
        self.write_z88elp(path / "z88elp.txt")
        self.write_z88int(path / "z88int.txt")
        self.write_z88mat(path / "z88mat.txt")
        self.write_z88man(path / "z88man.txt")
        self.write_z88dyn(path / "z88.dyn")

    def read_z88o2(self, path: Path) -> pd.DataFrame:
        return pd.read_csv(
            path,
            usecols=[1, 2, 3],
            names=["u1", "u2", "u3"],
            skiprows=5,
            nrows=self.n_nodes,
            delim_whitespace=True,
        )

    def read_z88o3(self, path: Path) -> pd.DataFrame:
        data = []

        with path.open() as f:
            for _ in range(7):
                f.readline()

            for i in range(self.n_elements):
                for _ in range(2):
                    f.readline()

                while True:
                    line = f.readline().strip()
                    if len(line) == 0:
                        break

                    (
                        sig_xx,
                        sig_yy,
                        sig_zz,
                        tau_xy,
                        tau_yz,
                        tau_zx,
                    ) = map(float, line.split()[3:])

                    data.append([i, sig_xx, sig_yy, sig_zz, tau_xy, tau_yz, tau_zx])

        stress = pd.DataFrame(
            data,
            columns=[
                "element",
                "sig_xx",
                "sig_yy",
                "sig_zz",
                "tau_xy",
                "tau_yz",
                "tau_zx",
            ],
        )

        return stress

    def solve(self, executable: Path) -> Tuple[pd.DataFrame, pd.DataFrame]:
        with TemporaryDirectory() as tmpdirname:
            self.write(Path(tmpdirname))

            p = subprocess.run(
                [executable, "-c", "-sorcg"],
                cwd=tmpdirname,
                capture_output=True,
            )

            if (
                b"eps not reached, maxit reached, sorry!" in p.stdout
                or p.returncode != 0
            ):
                displacement = pd.DataFrame(
                    0.0, columns=["u1", "u2", "u3"], index=pd.RangeIndex(self.n_nodes)
                )

                stress = pd.DataFrame(
                    np.inf,
                    columns=[
                        "sig_xx",
                        "sig_yy",
                        "sig_zz",
                        "tau_xy",
                        "tau_yz",
                        "tau_zx",
                    ],
                    index=pd.RangeIndex(self.n_elements),
                )
                stress["element"] = stress.index
            else:
                displacement = self.read_z88o2(Path(tmpdirname) / "z88o2.txt")
                stress = self.read_z88o3(Path(tmpdirname) / "z88o3.txt")

        return displacement, stress

    def to_pyvista_mesh(self) -> pv.UnstructuredGrid:
        points = self.nodes[["x", "y", "z"]].to_numpy()

        element_nodes = np.vstack(
            self.elements[
                (self.elements.type == Z88ElementType.Tet4)
                | (self.elements.type == Z88ElementType.Tet10)
            ].nodes
        )[:, :4]

        cells = np.hstack(
            [
                np.full(element_nodes.shape[0], 4, dtype=np.int64).reshape(-1, 1),
                element_nodes,
            ]
        ).flatten()

        celltypes = np.full(element_nodes.shape[0], vtk.VTK_TETRA, dtype=np.int64)

        return pv.UnstructuredGrid(cells, celltypes, points)


def main():
    parser = argparse.ArgumentParser(
        description="Automated Simulation of Metal FFF 3D Printing Debinding Process."
    )
    parser.add_argument(
        "input_file", type=str, help="Path to input file (STEP/STL), modelled in [mm]."
    )
    parser.add_argument(
        "-e",
        "--executable",
        type=str,
        default="z88r",
        help="Path to Z88R solver executable.",
    )
    parser.add_argument(
        "-x",
        "--scale-factor-xy",
        type=float,
        default=1.2,
        help="Scaling factor in building plane [-].",
    )
    parser.add_argument(
        "-z",
        "--scale-factor-z",
        type=float,
        default=1.27,
        help="Scaling factor in stacking direction [-].",
    )
    parser.add_argument(
        "-s", "--element-size", type=float, default=2.0, help="Mesh element size [mm]."
    )
    parser.add_argument(
        "-o",
        "--element-order",
        type=int,
        default=2,
        choices=[1, 2],
        help="Shape function order of tetrahedrons.",
    )
    parser.add_argument(
        "-i",
        "--int-order",
        type=int,
        default=5,
        choices=[1, 4, 5],
        help="Number of integration points per tetrahedron.",
    )
    parser.add_argument(
        "-a",
        "--acceleration",
        type=float,
        default=9810.0,
        help="Acceleration due to gravity [mm/s^2].",
    )
    parser.add_argument(
        "-y",
        "--youngs-modulus",
        type=float,
        default=210000,
        help="Young's modulus of the material [N/mm^2].",
    )
    parser.add_argument(
        "-p",
        "--poissons-ratio",
        type=float,
        default=0.4,
        help="Poisson's ratio of the material [-]",
    )
    parser.add_argument(
        "-d",
        "--density",
        type=float,
        default=4.7e-9,
        help="Density of the material [t/mm^3].",
    )
    parser.add_argument(
        "-f",
        "--max-tensile-stress-xy",
        type=float,
        default=6.0e-3,
        help="Maximum tensile stress in printing plane [N/mm^2].",
    )
    parser.add_argument(
        "-b",
        "--max-compression-stress-xy",
        type=float,
        default=7.0e-3,
        help="Maximum compression stress in printing plane [N/mm^2].",
    )
    parser.add_argument(
        "-c",
        "--max-tensile-stress-z",
        type=float,
        default=0.5e-3,
        help="Maximum tensile stress in stacking direction [N/mm^2].",
    )
    parser.add_argument(
        "-w",
        "--max-compression-stress-z",
        type=float,
        default=7.0e-3,
        help="Maximum compression stress in stacking direction [N/mm^2].",
    )
    parser.add_argument(
        "-t",
        "--bc-tolerance",
        type=float,
        default=1.0e-1,
        help="Fixed boundary condition search tolerance [mm].",
    )
    args = parser.parse_args()

    args.max_tensile_stress_xy = abs(args.max_tensile_stress_xy)
    args.max_tensile_stress_z = abs(args.max_tensile_stress_z)
    args.max_compression_stress_xy = -abs(args.max_compression_stress_xy)
    args.max_compression_stress_z = -abs(args.max_compression_stress_z)

    # generate z88 project
    z88project_template = Z88Project.from_gmsh(
        Path(args.input_file),
        element_size=args.element_size,
        element_order=args.element_order,
    )

    # scale mesh
    z88project_template.transform_mesh(
        np.diag([args.scale_factor_xy, args.scale_factor_xy, args.scale_factor_z])
    )

    # set element integration order & material properties
    z88project_template.assign_integration_order(args.int_order, args.int_order)
    z88project_template.assign_material(args.youngs_modulus, args.poissons_ratio)

    # add gravity load
    z88project_template.assign_volume_force(args.density, args.acceleration)

    # transformation matrices
    matrices = [
        np.eye(3),
        Rotation.from_rotvec(np.pi / 2 * np.array([1, 0, 0])).as_matrix(),
        Rotation.from_rotvec(np.pi * np.array([1, 0, 0])).as_matrix(),
        Rotation.from_rotvec(3 * np.pi / 2 * np.array([1, 0, 0])).as_matrix(),
        Rotation.from_rotvec(np.pi / 2 * np.array([0, 1, 0])).as_matrix(),
        Rotation.from_rotvec(-np.pi / 2 * np.array([0, 1, 0])).as_matrix(),
    ]

    # plotting
    plotter = pv.Plotter(shape=(2, 3))
    plotter.set_background("white")
    discrete_colormap = plt.cm.get_cmap("brg", 3)

    for i, matrix in enumerate(matrices):
        # create copy of z88 project
        z88project = copy.deepcopy(z88project_template)
        z88project.transform_mesh(matrix)

        # apply boundary conditions
        nodes_fixed = z88project.nodes[
            (z88project.nodes.z - z88project.nodes.z.min()).abs() <= args.bc_tolerance
        ].index

        for node in nodes_fixed:
            z88project.spcs.append(Z88SPC(node=node, dof=1, value=0.0))
            z88project.spcs.append(Z88SPC(node=node, dof=2, value=0.0))
            z88project.spcs.append(Z88SPC(node=node, dof=3, value=0.0))

        # solve project
        displacement, stress = z88project.solve(Path(args.executable))

        # get critical elements
        failure_tensile_xy = stress[
            (stress.sig_xx >= args.max_tensile_stress_xy)
            | (stress.sig_yy >= args.max_tensile_stress_xy)
        ].element.unique()

        failure_compression_xy = stress[
            (stress.sig_xx <= args.max_compression_stress_xy)
            | (stress.sig_yy <= args.max_compression_stress_xy)
        ].element.unique()

        failure_tensile_z = stress[
            (stress.sig_zz >= args.max_tensile_stress_z)
        ].element.unique()

        failure_compression_z = stress[
            (stress.sig_zz <= args.max_compression_stress_z)
        ].element.unique()

        # plot results
        plotter.subplot(i // 3, i % 3)

        element_colors = np.full(z88project.n_elements, 2)
        element_colors[failure_tensile_xy] = 1
        element_colors[failure_tensile_z] = 1
        element_colors[failure_compression_xy] = 0
        element_colors[failure_compression_z] = 0

        pyvista_mesh = z88project.to_pyvista_mesh()
        pyvista_mesh.cell_arrays["s"] = element_colors

        plotter.add_mesh(
            pyvista_mesh,
            scalars="s",
            show_edges=True,
            color="white",
            cmap=discrete_colormap,
        )
        plotter.remove_scalar_bar()

        plotter.add_mesh(
            pv.PolyData(z88project.nodes.loc[nodes_fixed, ["x", "y", "z"]].to_numpy()),
            point_size=10,
            render_points_as_spheres=True,
        )

        n_critical_elements = (
            len(failure_tensile_xy)
            + len(failure_compression_xy)
            + len(failure_tensile_z)
            + len(failure_compression_z)
        )

        print(f"Number of critical elements (variant {i}): {n_critical_elements}")

    plotter.show()


if __name__ == "__main__":
    main()
